/*
 Mastermind

 Concept: Krishna Lalith (https://www.hackster.io/krishnalalith/master-mind-game-5d2308)

 Case, Schematic & PCB design: jbrad2089@gmail.com
 Software: jbrad2089@gmail.com
 Took original and coded it to run on an Arduino UNO
 Replaced 4 Digit 7 Seg display with 1602 display
 Added player guess computer number mode

 5054 Unique combinations using 4 digits ranging from 0 to 9
 V3 - Worst guess is 9651 in 28 guesses
*/

#define AUTO 10000
#define SINGLE 9651
#define RELEASE 0

//Compile mode - AUTO => Test all possible numbers, SINGLE => Test single number, RELEASE => User interaction
#define COMPILE_MODE RELEASE

#if (COMPILE_MODE != RELEASE)
  #define DEBUG
#endif  

//include the library code:
#include <LiquidCrystal.h>
#include "buttons.h"

//LCD Screen
#define LCD_RS 8  //PB0 (14)
#define LCD_EN 9  //PB1 (15)
#define LCD_D4 10 //PB2 (16)
#define LCD_D5 11 //PB3 (17)
#define LCD_D6 12 //PB4 (18)
#define LCD_D7 13 //PB5 (19)

//Switch pins
#define COL_0 A0  //ADC0 (23)
#define COL_1 A1  //ADC1 (24)
#define COL_2 A2  //ADC2 (25)
#define COL_3 A3  //ADC3 (26)
#define ROW_0 4   //PD4 (6)
#define ROW_1 7   //PD7 (13)
#define ROW_2 6   //PD6 (12)
#define ROW_3 5   //PD5 (11)

//Unconnected Analog PIN for Random Number
#define RANDOM_SEED_PIN A4
 
//Initialize the LCD
LiquidCrystal lcd(LCD_RS, LCD_EN, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

//Extra character set codes
#define LCD_CORRECT 0
#define LCD_INCORRECT 1

//Create suit symbols character maps (5 columns x 8 rows).
byte correct[8] = {
  B00000,
  B01110,
  B11111,
  B11111,
  B11111,
  B01110,
  B00000,
  B00000,
};
byte incorrect[8] = {
  B00000,
  B01110,
  B10001,
  B10001,
  B10001,
  B01110,
  B00000,
  B00000,
};

//Switch values
#define SW_1B0C_7 0
#define SW_0B1C_8 1
#define SW_9 2
#define SW_NONE 3
#define SW_2B0C_4 4
#define SW_1B1C_5 5
#define SW_0B2C_6 6
#define SW_MATCH 7
#define SW_3B0C_0 8
#define SW_2B1C_1 9
#define SW_1B2C_2 10
#define SW_0B3C_3 11
#define SW_4B0C 12
#define SW_3B1C 13
#define SW_2B2C 14
#define SW_1B3C 15

#define INVALID 255

//Result mapping when user clicks button
//correct position - Bit 7, 6, 5, 4 (C code)
//wrong position - Bit 3, 2, 1, 0 (B Code)
const uint8_t results[] = { 
    (0 << 4) | 1, (1 << 4) | 0,      INVALID,            0,
    (0 << 4) | 2, (1 << 4) | 1, (2 << 4) | 0, (4 << 4) | 0,
    (0 << 4) | 3, (1 << 4) | 2, (2 << 4) | 1, (3 << 4) | 0,
    (0 << 4) | 4, (1 << 4) | 3, (2 << 4) | 2, (3 << 4) | 1
};

//These are special C codes that is used in the rules table
#define KNOWN 8
#define SOMEWHERE 7
#define ANYWHERE 6
#define NEVER 5

//This structure stores a guess or rule
struct guess
{
  char digits[4];
  uint8_t result;
  char special;
};

//Records the guesses and result from the guesses made
#define MAX_GUESSES 50
guess history[MAX_GUESSES];
uint8_t historyCount = 0;
#define MAX_RULES 50
guess rules[MAX_RULES];
uint8_t rulesCount = 0;

char known[4];
uint8_t knownDigits;

uint16_t nextGuess = 0;
String digits;
String myDigits;

enum GameModeEnum { MODE_UNKNOWN, MODE_I_GUESS, MODE_U_GUESS };

//Define callback for input
int (*pInputFunction)(String digits);
void (*pOutputFunction)(String digits);

void setup() 
{
  #ifdef DEBUG
    Serial.begin(115200);
  #endif

  //Define all the new card characters
  lcd.createChar(LCD_CORRECT, correct);
  lcd.createChar(LCD_INCORRECT, incorrect);

  //Setup display
  lcd.begin(16, 2);
  lcd.clear();
  lcd.setCursor(0,0);
  //Setup switches
  initButtons(ROW_0, ROW_1, ROW_2, ROW_3, COL_0, COL_1, COL_2, COL_3);

  #if (COMPILE_MODE == RELEASE)
    showSplashScreen();
  #endif
}

void loop() 
{
  
#if (COMPILE_MODE == AUTO)

  int maxNumberOfGuesses = 0;
  String worstDigits;
  
  for (int test = 123; test < 10000; test++)
  {
    myDigits = ((test < 1000) ? "0" : "") + String(test);
    if (myDigits[0] != myDigits[1] && myDigits[0] != myDigits[2] && myDigits[0] != myDigits[3] && myDigits[1] != myDigits[2] && myDigits[1] != myDigits[3] && myDigits[2] != myDigits[3])
    {
      lcd.clear();
      lcd.setCursor(0,0);
      lcd.print("My Guess is " + String(myDigits));
      //Serial.println("Trying to guess " + String(myDigits));
      int numberOfGuesses = computerGuessPlayer(&getAutoResponse, &displayComputerGuess);
      if (numberOfGuesses == 0)
      {
        #ifdef DEBUG
          Serial.println(" - Can't guess " + String(myDigits) + ", history depth is " + String(historyCount) + ", rule depth is " + String(rulesCount));
        #endif
      }
      else if (numberOfGuesses > maxNumberOfGuesses)
      {
        maxNumberOfGuesses = numberOfGuesses;
        worstDigits = myDigits;
        #ifdef DEBUG
          Serial.println(" - Worst guess so far is " + String(worstDigits) + " in " + String(maxNumberOfGuesses) + " guesses");
        #endif
      }
    }
  }
  #ifdef DEBUG  
    Serial.println("Worst guess " + String(worstDigits) + " in " + String(maxNumberOfGuesses) + " guesses");
  #endif
  
#elif (COMPILE_MODE != RELEASE)

  int maxNumberOfGuesses = 0;
  String worstDigits;
  
  int test = SINGLE;
  myDigits = ((test < 1000) ? "0" : "") + String(test);
  if (myDigits[0] != myDigits[1] && myDigits[0] != myDigits[2] && myDigits[0] != myDigits[3] && myDigits[1] != myDigits[2] && myDigits[1] != myDigits[3] && myDigits[2] != myDigits[3])
  {
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("My Guess is " + String(myDigits));
    
    #ifdef DEBUG  
      Serial.println("Trying to guess " + String(myDigits));
    #endif
    int numberOfGuesses = computerGuessPlayer(&getAutoResponse, &displayComputerGuess);
    if (numberOfGuesses == 0)
    {
      #ifdef DEBUG  
        Serial.println("Can't guess " + String(myDigits) + ", history depth is " + String(historyCount) + ", rule depth is " + String(rulesCount));
      #endif
    }
    else if (numberOfGuesses > maxNumberOfGuesses)
    {
      maxNumberOfGuesses = numberOfGuesses;
      worstDigits = myDigits;
    }
  }
  #ifdef DEBUG  
    Serial.println("Worst guess " + String(worstDigits) + " in " + String(maxNumberOfGuesses) + " guesses, history depth is " + String(historyCount) + ", rule depth is " + String(rulesCount));
  #endif
  while (true)
  {
    delay(1000);
  }
  
#else

  //First find out what the game is
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Who guesses?");
  lcd.setCursor(0, 1);
  lcd.print("[1] You   [2] Me");
    
  GameModeEnum gameMode = MODE_UNKNOWN;
  while (gameMode == MODE_UNKNOWN)
  {
    int key = getUserNumber(false);
    switch (key)
    {
      case 1: gameMode = MODE_U_GUESS; break;
      case 2: gameMode = MODE_I_GUESS; break;
    }
  }

  randomSeed(millis());

  //Play game
  switch (gameMode)
  {
    case MODE_I_GUESS: computerGuesses(); break;
    case MODE_U_GUESS: playerGuesses(); break;
  }


#endif
}

//-------------------------------------------------------------------------------------------------

//Play a round where the computer guesses the players number
void computerGuesses()
{
  //Get player to enter their number so they don't forget it
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Enter 4 digits:");
  lcd.setCursor(0, 1);
  lcd.blink();
  String number = "";
  while (number.length() < 4)
  {
    int key = getUserNumber(false);
    if (key != -1)
    {
      char digit = (key + 48);
      //Test if key already pressed
      int i = 0;
      while (i < number.length() && number[i] != digit)
      {
        i++;
      }
      if (i >= number.length())
      {
        lcd.print(digit);
        number += digit;
      }
    }
  }
  lcd.clear();
  lcd.noBlink();
  lcd.setCursor(0, 0);
  lcd.print("Number is " + number);
  int numberOfGuesses = computerGuessPlayer(&getUserResponse, &displayComputerGuess);
  lcd.setCursor(0, 1);
  lcd.print("                ");
  lcd.setCursor(0, 1);
  if (numberOfGuesses > 0)
  {
    lcd.print("Guessed in " + String(numberOfGuesses));
  }
  else
  {
    lcd.print("You cheated!");
  }
  waitOnKeyPress();
}

//-------------------------------------------------------------------------------------------------

//Play a round where the player guesses the computers number
void playerGuesses()
{
  myDigits = getRandomNumber();
  int numberOfGuesses = 0;
  String guessNumber;
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Enter guesses:");
  bool solved = false;
  while (!solved && numberOfGuesses < 99)
  {
    numberOfGuesses++;
    guessNumber = (numberOfGuesses < 10 ? "0" : "") + String(numberOfGuesses) + ": ";
    lcd.setCursor(0, 1);
    lcd.print(guessNumber + "            ");
    lcd.setCursor(4, 1);
    lcd.blink();
    String number = "";
    while (number.length() < 4)
    {
      int key = getUserNumber(true);
      if (key == 10)
      {
        number = myDigits;
        lcd.setCursor(0, 1);
        lcd.print("  ** MATCH **   ");
      }
      else if (key != -1)
      {
        char digit = (key + 48);
        //Test if key already pressed
        int i = 0;
        while (i < number.length() && number[i] != digit)
        {
          i++;
        }
        if (i >= number.length())
        {
          lcd.print(digit);
          number += digit;
        }
      }
    }
    lcd.noBlink();

    //Work out results
    int c = 0;
    int b = 0;
    for (int j = 0; j < 4; j++)
    {
      for (int i = 0; i < 4; i++)
      {
        if (number[i] == myDigits[j])
        {
          if (i == j)
          {
            c++;
          }
          else
          {
            b++;
          }
        }
      }
    }

    //Display Result
    lcd.setCursor(0, 0);
    lcd.print(guessNumber + number + " -      ");
    lcd.setCursor(11, 0);
    for (int i = 0; i < c; i++)
    {
      lcd.write((byte)LCD_CORRECT);
    }
    for (int i = 0; i < b; i++)
    {
      lcd.write((byte)LCD_INCORRECT);
    }
    if (c == 0 && b == 0)
    {
      lcd.print("----");
    }

    solved = (c == 4);
  }
  waitOnKeyPress();
}

//-------------------------------------------------------------------------------------------------

//Wait until any key is pressed
void waitOnKeyPress()
{
  //Get Result
  int button = NO_BUTTON;
  while (button == NO_BUTTON) 
  {
    button = buttonPressed();
    delay(10);
  }
  
}

//-------------------------------------------------------------------------------------------------

//Get the response from the player for the guess made by the computer
// digits - Guess made
// Returns button that matches the response
int getAutoResponse(String digits)
{
  //Count digits in the correct position
  int button = SW_9;
  int c = 0;
  int b = 0;
  for (int j = 0; j < 4; j++)
  {
    for (int i = 0; i < 4; i++)
    {
      if (digits[i] == myDigits[j])
      {
        if (i == j)
        {
          c++;
        }
        else
        {
          b++;
        }
      }
    }
  }
  uint8_t result = c << 4 | b;
  for (int i = 0; i < 16; i++)
  {
    if (results[i] == result)
    {
      button = i;
      break;
    }
  }

  //Serial.println("Guess " + String(digits[0]) + String(digits[1]) + String(digits[2]) + String(digits[3]) + ", c = " + String(c) + ", b = " + String(b));
  return button;
}

//-------------------------------------------------------------------------------------------------

//Get the response from the player for the guess made by the computer
// digits - Guess made
// Returns button that matches the response
int getUserResponse(String digits)
{
  //Get Result
  int button = NO_BUTTON;
  while (button == NO_BUTTON || results[button] == INVALID) 
  {
    button = buttonPressed();
    delay(10);
  }
  return button;
}

//-------------------------------------------------------------------------------------------------

//Get the response from the player for a number
// terminateKey - Whether the MATCH key is valid (returns 10 if it is)
// Returns the digit key pressed (0..9) or -1
int getUserNumber(bool terminateKey)
{
  int result = -1;
  int button = NO_BUTTON;
  while (button == NO_BUTTON) 
  {
    button = buttonPressed();
    delay(10);
  }
  switch (button)
  {
    case SW_3B0C_0: result = 0; break;
    case SW_2B1C_1: result = 1; break;
    case SW_1B2C_2: result = 2; break;
    case SW_0B3C_3: result = 3; break;
    case SW_2B0C_4: result = 4; break;
    case SW_1B1C_5: result = 5; break;
    case SW_0B2C_6: result = 6; break;
    case SW_1B0C_7: result = 7; break;
    case SW_0B1C_8: result = 8; break;
    case SW_9: result = 9; break;
    case SW_MATCH: 
      if (terminateKey)
      {
        result = 10;
      }
      break;
  }
  return result;
}

//-------------------------------------------------------------------------------------------------

//Display the guess that the computer has made
// digits - Guess to display
void displayComputerGuess(String digits)
{
  #if (COMPILE_MODE != AUTO)
    lcd.setCursor(0, 1);
    lcd.print("I guess " + digits);
  #endif
}

//-------------------------------------------------------------------------------------------------

//Return a valid random 4 digit number
String getRandomNumber()
{
  String number = "";
  for (int j = 0; j < 4; j++)
  {
    bool added = false;
    while (!added)
    {    
      char digit = random((j == 0) ? 1 : 0, 10) + 48;
      //Test if we already have this digit
      int i = 0;
      while (i < number.length() && number[i] != digit)
      {
        i++;
      }
      if (i >= number.length())
      {
        number += digit;
        added = true;
      }
    }
  }
  return number;    
}

//-------------------------------------------------------------------------------------------------

//Start guessing the player's number
// pInputFunction - Function invoked to get the response to the guess
// pOutputFunction - Function invoked to show what guess is being made
// Returns the number of guesses made to find the number
int computerGuessPlayer(int (*pInputFunction)(String digits), void (*pOutputFunction)(String digits))
{
  //Clear known array
  knownDigits = 0;
  historyCount = 0;
  rulesCount = 0;
  
  int finalGuesses = 0;
  int numberOfGuesses = 0;
  for (int i = 0; i < 4; i++)
  {
    known[i] = '\0';
  }
  for (nextGuess = 123; nextGuess < 10000; nextGuess++)
  {
    //Break down guess into individual digits
    digits = ((nextGuess < 1000) ? "0" : "") + String(nextGuess);

    //Continue if no duplicates
    if (digits[0] != digits[1] && digits[0] != digits[2] && digits[0] != digits[3] && digits[1] != digits[2] && digits[1] != digits[3] && digits[2] != digits[3])
    {
      if (passRuleCheck(digits))
      {
        //Display this guess
        #if (COMPILE_MODE == SINGLE)
          #ifdef DEBUG  
            Serial.print(" - guess " + String(digits[0]) + String(digits[1]) + String(digits[2]) + String(digits[3]));
          #endif          
        #endif
        pOutputFunction(digits);
        numberOfGuesses++;

        //Find out results
        uint16_t button = pInputFunction(digits);
        if (button == SW_MATCH)
        {
          #if (COMPILE_MODE == SINGLE)
            #ifdef DEBUG  
              Serial.println(", MATCH");
            #endif
          #endif
          //Found solution.
          finalGuesses = numberOfGuesses;
          break;
        }
        else
        {
          uint8_t result = results[button];
          uint8_t c = result >> 4;
          uint8_t b = result & 0x0F;

          #if (COMPILE_MODE == SINGLE) 
            #ifdef DEBUG  
              Serial.println(", c=" + String(c) + ", b=" + String(b));
            #endif
          #endif
          
          //If we only have known digits, add new rule to block all the other digits
          if (c == knownDigits && b == 0)
          {
            //Digits other than those known are invalid
            for (int i = 0; i < 4; i++)
            {
              if (known[i] == '\0')
                insertRule(digits, (NEVER << 4) | i);
            }
          }
          
          //Check to see if this was worse than our last guess
          if (historyCount > 0)
          {
            uint8_t lastPos = historyCount - 1;
            uint8_t last_c = history[lastPos].result >> 4;
            uint8_t last_b = history[lastPos].result & 0x0F;
            //We can tell a lot if only digit changed so test for this
            int pos = -1;
            int cnt = 0;
            for (int i = 0; i < 4; i++)
            {
              if (history[lastPos].digits[i] == digits[i])
              {
                cnt++;
              }
              else
              {
                pos = i;
              }
            }

            /* Fails on 0134
            if ((last_c == 3 && last_b == 0 && (b + c) == 3) || (c == 3 && b == 0 && (last_b + last_c) == 3))
            {
              //There were three in the correct position now and some in the wrong position. 
              //Just find the matching digits and their old position
              int noMatchCnt = 0;
              int noMatchPos = 0;
              for (int j = 0; j < 4; j++)
              {
                char t = history[lastPos].digits[j];
                bool found = false;
                for (int i = 0; i < 4; i++)
                {
                  if (t == digits[i])
                  {
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  noMatchCnt++;
                  noMatchPos = j;
                }
              }
              if (noMatchCnt == 1)
              {
                //All other positions are known
                for (int i = 0; i < 4; i++)
                {
                  if (i != noMatchPos)
                  {
                    if (c == 3)
                    {
                      known[pos] = digits[i];
                      knownDigits++;
                      insertRule(digits, (KNOWN << 4) | i);
                    }
                    else
                    {
                      known[pos] = history[lastPos].digits[i];
                      knownDigits++;
                      insertRule(history[lastPos].digits, (KNOWN << 4) | i);
                    }
                  }
                }
              }
            }
            else */ 
            if (last_c == 2 && knownDigits == 2)
            {
              if (b == 2)
              {
                //Serial.println("Rule last_c == 2 && knownDigits == 2 && b == 2");
                //Know two digits and the other two are correct but in the wrong order so swap them
                int8_t pos_1 = -1;
                int8_t pos_2 = -1;
                for (int i = 0; i < 4; i++)
                {
                  if (known[i] == '\0')
                  {
                    if (pos_1 == -1)
                    {
                      pos_1 = i;
                    }
                    else
                    {
                      pos_2 = i;
                    }
                  }
                }
                //Swap them
                char t = digits[pos_1];
                digits[pos_1] = digits[pos_2];
                digits[pos_2] = t;
                //Record what we know
                known[pos_1] = digits[pos_1];
                knownDigits++;
                known[pos_2] = digits[pos_2];
                knownDigits++;
                //add the rules
                insertRule(digits, (KNOWN << 4) | pos_1);
                insertRule(digits, (KNOWN << 4) | pos_2);
                button = SW_MATCH;
                result = results[button];
              }
              else if (b == 1 && last_b == 1)
              {
                //Check to see if only one digit changed. If so, other digit is correct but in the wrong position
                int8_t notChanged = -1;
                int8_t changed = -1;
                int cntChanged = 0;
                for (int i = 0; i < 4; i++)
                {
                  if (known[i] == '\0')
                  {
                    if (history[lastPos].digits[i] == digits[i])
                    {
                      notChanged = i;
                    }
                    else
                    {
                      cntChanged++;
                      changed = i;
                    }
                  }
                  else if (known[i] != history[lastPos].digits[i] || known[i] != digits[i])
                  {
                    cntChanged++;
                  }
                }
                if (cntChanged == 1 && changed != -1)
                {
                  //Serial.println("Changed = " + String(changed) + ", notChanged = " + String(notChanged));
                  //swap digits
                  char t = digits[notChanged];
                  digits[notChanged] = digits[changed];
                  digits[changed] = t;
                  //Record what we know
                  known[changed] = digits[changed];
                  knownDigits++;
                  //Create a new rule at highest level since we know that this digit is in the correct position
                  insertRule(digits, (KNOWN << 4) | changed );
                  button = SW_0B3C_3;
                  result = results[button];
                }
              }
            }
            else if (b == 4)
            {
              //All present and all are in the wrong position
              //Serial.println("    - test (b == 4)");
              insertRule(digits, (SOMEWHERE << 4) | 0);
              insertRule(digits, (SOMEWHERE << 4) | 1);
              insertRule(digits, (SOMEWHERE << 4) | 2);
              insertRule(digits, (SOMEWHERE << 4) | 3);
            }
            else if ((b + c) == 4)
            {
              //All present but some are in the wrong position
              //Serial.println("    - test (b + c) == 4");
              insertRule(digits, (ANYWHERE << 4) | 0);
              insertRule(digits, (ANYWHERE << 4) | 1);
              insertRule(digits, (ANYWHERE << 4) | 2);
              insertRule(digits, (ANYWHERE << 4) | 3);
            }
            else if (cnt == 3)
            {
              //Serial.println("    - only one changed");
              //Three numbers haven't changed. This can tell a lot
              if (last_c > c)
              {
                //Serial.println("    - test last_c > c");
                //last_c is in the solution 
                known[pos] = history[lastPos].digits[pos];
                knownDigits++;
                insertRule(history[lastPos].digits, (KNOWN << 4) | pos );
              }
              else if (last_c < c)
              {
                //Serial.println("    - test last_c < c");
                //c is in the solution 
                //Record what we know
                known[pos] = digits[pos];
                knownDigits++;
                insertRule(digits, (KNOWN << 4) | pos );
              }
              else if (last_b > b)
              {
                //Serial.println("    - test last_c == c && last_b > b");
                //b is not in the solution
                insertRule(digits, (NEVER << 4) | pos);
              }
              else if (last_b < b)
              {
                //Serial.println("    - test last_c == c && last_b < b");
                //b is in the solution but not in the correct position
                insertRule(digits, (SOMEWHERE << 4) | pos);
              }
              else if (b == 0)
              {
                //Serial.println("    - test last_c == c && b == 0");
                //neither last_b or b is present
                uint8_t rule = (NEVER << 4) | pos;
                insertRule(history[lastPos].digits, rule);
                insertRule(digits, rule);
              }
              /* fails 3456
              else if (c > 0 && b == 1 && historyCount > 1)
              {
                //Serial.println("    - test last_c == c && b == 1 && historyCount > 1");
                //only one digit changed but because b = 1, can't be sure that b refers to the same digit. However
                //if the previous to last matches, then all three changed digits are not present
                uint8_t lastLastPos = historyCount - 2;
                uint8_t lastLast_c = history[lastLastPos].result >> 4;
                uint8_t lastLast_b = history[lastLastPos].result & 0x0F;
                if (lastLast_c == c && lastLast_b == b)
                {
                  cnt = 0;
                  for (int i = 0; i < 4; i++)
                  {
                    if (history[lastPos].digits[i] == digits[i] && history[lastLastPos].digits[i] == digits[i])
                    {
                      cnt++;
                    }
                    else
                    {
                      pos = i;
                    }
                  }
                  if (cnt == 3)
                  {
                    //No change in results across three tries where the same digit has changed 3 times
                    uint8_t rule = (NEVER << 4) | pos;
                    insertRule(history[lastLastPos].digits, rule);
                    insertRule(history[lastPos].digits, rule);
                    insertRule(digits, rule);
                  }
                }
              }
              */
              else if (b <= 2 && historyCount > 2)
              {
                //Serial.println("    - test last_c == c && b == 2 && historyCount > 2");
                //only one digit changed but because b = 2, can't be sure that b refers to the same digit. However
                //if the previous to last matches and previous of previous to last matches, then all four changed digits are not present
                uint8_t lastLastLastPos = historyCount - 3;
                uint8_t lastLastPos = historyCount - 2;
                uint8_t lastLast_c = history[lastLastPos].result >> 4;
                uint8_t lastLast_b = history[lastLastPos].result & 0x0F;
                uint8_t lastLastLast_c = history[lastLastLastPos].result >> 4;
                uint8_t lastLastLast_b = history[lastLastLastPos].result & 0x0F;
                if (lastLast_c == c && lastLast_b == b && lastLastLast_c == c && lastLastLast_b == b)
                {
                  cnt = 0;
                  for (int i = 0; i < 4; i++)
                  {
                    if (history[lastPos].digits[i] == digits[i] && history[lastLastPos].digits[i] == digits[i] && history[lastLastLastPos].digits[i] == digits[i])
                    {
                      cnt++;
                    }
                    else
                    {
                      pos = i;
                    }
                  }
                  if (cnt == 3)
                  {
                    //No change in results across four tries where the same digit has changed 4 times
                    uint8_t rule = (NEVER << 4) | pos;
                    insertRule(history[lastLastLastPos].digits, rule);
                    insertRule(history[lastLastPos].digits, rule);
                    insertRule(history[lastPos].digits, rule);
                    insertRule(digits, rule);
                  }
                }
              }
            }
          }
          insertHistory(digits, result);
        }        
      }
    }
  }
  return finalGuesses;
}

//-------------------------------------------------------------------------------------------------

//Test whether guess passes existing rules
// digits - The current guess to try
// Returns true if guess is OK to try out
bool passRuleCheck(String digits)
{
  bool pass = true;

  //Start by counting the number of digits that match the guess we did before
  for (int r = 0; r < rulesCount; r++)
  {
    uint8_t c = rules[r].result >> 4;
    uint8_t b = rules[r].result & 0x0F;

    //Serial.println("Guess " + String(digits[0]) + String(digits[1]) + String(digits[2]) + String(digits[3]) + ", Rules " + String(rules[r].digits[0]) + String(rules[r].digits[1]) + String(rules[r].digits[2]) + String(rules[r].digits[3]) + ", result is (" + String(c) + "," + String(b) + ")");

    if (c == KNOWN)
    {
      //Special rule inserted when a digit is known
      if (rules[r].digits[b] != digits[b])
      {
        pass = false;
        break;
      }
    }
    
    else if (c == SOMEWHERE)
    {
      pass = false;
      //Special rule inserted for known digits somewhere in number but not at the position b
      for (int i = 0; i < 4; i++)
      {
        if (rules[r].digits[b] == digits[i] && b != i)
        {
          pass = true;
          break;
        }
      }
      if (!pass)
      {
        break;
      }
    }

    else if (c == ANYWHERE)
    {
      pass = false;
      //Special rule inserted for known digits somewhere in number at any position
      for (int i = 0; i < 4; i++)
      {
        if (rules[r].digits[b] == digits[i])
        {
          pass = true;
          break;
        }
      }
      if (!pass)
      {
        break;
      }
    }

    else if (c == NEVER)
    {
      //Special rule inserted for digits known not to be in number
      for (int i = 0; i < 4; i++)
      {
        if (rules[r].digits[b] == digits[i])
        {
          pass = false;
          break;
        }
      }
      if (!pass)
      {
        break;
      }
    }

    else
    {
      pass = passHistoryCheck(digits);      
    }

    //Don't keep testing if we find one combination that fails
    if (!pass)
      break;
  }
 
  return pass;
}

//-------------------------------------------------------------------------------------------------

//Test whether guess passes existing rules
// digits - The current guess to try
// Returns true if guess is OK to try out
bool passHistoryCheck(String digits)
{
  bool pass = true;

  //Start by counting the number of digits that match the guess we did before
  for (int h = 0; h < historyCount; h++)
  {
    //Serial.println("Guess " + String(digits[0]) + String(digits[1]) + String(digits[2]) + String(digits[3]) + ", History " + String(history[h].digits[0]) + String(history[h].digits[1]) + String(history[h].digits[2]) + String(history[h].digits[3]) + ", result is (" + String(c) + "," + String(b) + ")");

    //Count the number of times each digit appears
    uint8_t sum = 0;
    for (int j = 0; j < 4; j++)
    {
      for (int i = 0; i < 4; i++)
      {
        if (history[h].digits[j] == digits[i])
        {
          sum++;
        }
      }
    }
    //Serial.println(", SUM is " + String(sum));

    uint8_t c = history[h].result >> 4;
    uint8_t b = history[h].result & 0x0F;
    switch (c + b)
    {
      case 0:
        //None of the digits are present
        pass = (sum == 0);
        break;

      case 1:
        //One of the digits is present
        pass = (sum >= 1);
        break;
        
      case 2:
        //Two of the digits are present
        pass = (sum >= 2);
        break;
        
      case 3:
        //Three of the digits are present
        pass = (sum >= 3);
        break;
        
      case 4:
        //Four of the digits are present
        pass = (sum >= 4);
        break;
      
    }

    //Don't keep testing if we find one combination that fails
    if (!pass)
      break;
  }
 
  return pass;
}

//-------------------------------------------------------------------------------------------------

//Insert guess in history array
//followed by the number of correct digits in the wrong position
// digits - The guess this items relates to
// button - The button pressed relating to the guess
// result - The results value (Bit 3-5 is Digits in correct position, Bits 0-2 is Digits correct but in wrong position
// Returns position of item inserted
uint8_t insertHistory(String digits, int8_t result)
{
  //Store the rule
  int pos = historyCount;
  for (int i = 0; i < 4; i++)
  {
    history[pos].digits[i] = digits[i];
  }
  history[pos].result = result;
  historyCount++;
  return pos;
}

//-------------------------------------------------------------------------------------------------

//Insert new rule into a sorted rule array ignoring duplicates
// digits - The guess this items relates to
// result - The results value (Bit 3-5 is KNOWN. SOMEWHERE or NEVER, Bits 0-2 is position in digits
// Returns position of item inserted
uint8_t insertRule(String digits, uint8_t result)
{
  //First get out the digit this rule refers to
  char c = result >> 4;
  char special = digits[result & 0x0F];
  //First test if rule is already present
  int pos = 0;
  while (pos < rulesCount && (rules[pos].result > result || (rules[pos].result == result && rules[pos].special > special)))
  {
    pos++;
  }
  if (pos >= rulesCount || rules[pos].result != result || rules[pos].special != special)
  {
    //Not in list so insert at pos and move rest up
    int mov = rulesCount;
    while (mov > pos)
    {
      for (int i = 0; i < 4; i++)
      {
        rules[mov].digits[i] = rules[mov - 1].digits[i];
      }
      rules[mov].result = rules[mov - 1].result;
      rules[mov].special = rules[mov - 1].special;
      mov--;
    }
    
    //Insert new one at pos
    #if (COMPILE_MODE == SINGLE)
      #ifdef DEBUG  
        Serial.println("   - rule: " + String(digits[0]) + String(digits[1]) + String(digits[2]) + String(digits[3]) + ", result is (" + String(result >> 4) + "," + String(result & 0x0F) + ")");
      #endif
    #endif
    
    //Store the rule, 
    for (int i = 0; i < 4; i++)
    {
      rules[pos].digits[i] = digits[i];
    }
    rules[pos].result = result;
    rules[pos].special = special;
    rulesCount++;
  }
  return pos;
}

//-----------------------------------------------------------------

//Show the startup screen
void showSplashScreen()
{
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("** MASTERMIND **");
  lcd.setCursor(6, 1);
  lcd.write((byte)LCD_INCORRECT);
  lcd.write((byte)LCD_INCORRECT);
  lcd.write((byte)LCD_INCORRECT);
  lcd.write((byte)LCD_INCORRECT);
  delay(1000);

  for (int i = 6; i < 10; i++)
  {
    lcd.setCursor(i, 1);
    lcd.write((byte)LCD_CORRECT);
    delay(1000);
  }
  
  int animDelay = 10;
  while (animDelay < 31)
  {
    animDelay = animDelay + 1;
    if ((animDelay % 2) > 0)
      lcd.display();
    else
      lcd.noDisplay();
    delay((31 - animDelay) * (31 - animDelay));
  };
  delay(1000);
};
